require 'nexpose_thycotic/utilities/nx_logger'
require 'set'
# Required for use of v3 endpoints for patch
require 'base64'
require 'net/http'
require 'json'

module NexposeThycotic
  class NexposeOperations
    require 'nexpose'
    include Nexpose
    attr_reader :nsc

    def initialize(ip, user, pass, port = 3780)
      @nsc = Connection.new(ip, user, pass, port)
      @nsc.login

      @logger = NexposeThycotic::NxLogger.instance
      @logger.on_connect(ip, port, @nsc.session_id, '{}')

      @insightvm_client = InsightvmClient.new(ip, port, user, pass, @logger)
    end

    def get_site(site_id)
      Site.load(@nsc, site_id)
    end

    def get_devices_from_site(site_id)
      @nsc.list_devices(site_id)
    end

    def get_device_addresses(site_id)
      site = get_site(site_id)
      included_targets = site.included_addresses

      # Get a list of known IP addresses from a site
      devices = get_devices_from_site(site_id)
      device_ips = devices.map { |d| d.address }

      hosts = []
      ips = device_ips.to_set
      included_targets.each do |address|
        if address.instance_of?(Nexpose::HostName)
          hosts << address.host
        elsif address.to.nil?
          ips.add(address.from)
        end
      end

      @logger.debug("Discovered #{hosts.count} hosts and #{ips.count} IPs.")
      ips.to_a + hosts
    end

    def get_existing_credentials(site_id)
      # @insightvm_client.get_site_credentials(site_id)
      site = get_site(site_id)
      site.site_credentials
    end

    def save_site(site_id, credentials, overwrite=true)
      creds_bulk_update = []
      credentials.each do |cred|
        # Skipping or deleting credential as no password exists; this could occur due to no password returned by
        # Thycotic or because the credential was returned as an existing credential for the site
        if cred.password.nil?
          # Credentials to delete (in overwrite mode); those with no password and an id
          if overwrite && cred.id != -1
            @insightvm_client.delete_site_credential(site_id, cred.id)
          end

          # Move on to next credential
          next
        end

        # Convert nexpose cred to v3 json payload
        c = to_v3_credential(cred)

        if c["id"] == -1
          # Add individual if new
          @insightvm_client.create_site_credential(site_id, c)
          @logger.debug("Creating Site #{site_id} credential #{c["name"]}")
        else
          creds_bulk_update.push(c)
        end
      end

      # Bulk update any credentials that already exist to be more efficient
      @insightvm_client.update_site_credentials(site_id, creds_bulk_update) unless creds_bulk_update.empty?
    end

    def delete_site_credentials(site_id)
      save_site(site_id, [])
    end

    def start_scan(site_id)
      site = get_site(site_id)
      site.scan(@nsc)
    end

    def scan_status(scan_id)
      @nsc.scan_status(scan_id)
    end

    private

    # Convert Nexpose client Site Credential to v3 Credential
    def to_v3_credential(credential)
      cred = {
          "account": {
              "service": credential.service,
              "username": credential.user_name,
              "password": credential.password
          },
          "name": credential.name,
          "description": credential.description,
          "enabled": true,
          "hostRestriction": credential.host_restriction,
      }
      # Set credential ID so it can be updated instead of a new one created
      unless credential.id.nil?
        cred["id"] = credential.id
      end
      cred
    end
  end

  class InsightvmClient
    def initialize(host, port, user, pass, logger)
      @host = host
      @port = port
      @auth_header = {'Authorization': "Basic #{Base64.strict_encode64("#{user}:#{pass}")}"}
      @logger = logger
    end

    def request(uri, request)
      # Set up http client
      http = Net::HTTP.new(uri.host, uri.port)

      # Ignore certificate verification if using https; this is consistent with the rest of the integration
      if uri.scheme == "https"
        http.use_ssl = true
        http.verify_mode = OpenSSL::SSL::VERIFY_NONE
      end

      response = http.request(request)
      case response
      when Net::HTTPOK, Net::HTTPCreated, Net::HTTPNoContent
        response.body
      else
        raise Nexpose::AJAX.get_api_error(request, response)
      end
    end

    def get(uri, content_type = 'application/json')
      request = Net::HTTP::Get.new(uri.request_uri, @auth_header)
      request.set_content_type(content_type)

      request(uri, request)
    end

    def post(uri, payload = nil, content_type = 'application/json')
      request = Net::HTTP::Post.new(uri.request_uri, @auth_header)
      request.set_content_type(content_type)
      request.body = payload.to_json if payload

      request(uri, request)
    end

    def put(uri, payload = nil, content_type = 'application/json')
      request = Net::HTTP::Put.new(uri.request_uri, @auth_header)
      request.set_content_type(content_type)
      request.body = payload.to_json if payload

      request(uri, request)
    end

    def delete(uri, content_type = 'application/json')
      request = Net::HTTP::Delete.new(uri.request_uri, @auth_header)
      request.set_content_type(content_type)

      request(uri, request)
    end

    def get_site_credentials(site_id)
      uri = URI("https://#{@host}:#{@port}/api/3/sites/#{site_id}/site_credentials")
      begin
        credentials = get(uri)
        creds = JSON.parse(credentials)

        return creds["resources"]
      rescue Exception => e
        @logger.error("Failed to retrieve Site #{site_id} credentials; #{e}")
      end
    end

    def delete_site_credential(site_id, credential_id)
      uri = URI("https://#{@host}:#{@port}/api/3/sites/#{site_id}/site_credentials/#{credential_id}")

      begin
        delete(uri)
        @logger.info("Site #{site_id} credential removed successfully")
      rescue Exception => e
        @logger.error("Failed to remove Site #{site_id} credential; #{e}")
      end
    end

    def delete_site_credentials(site_id)
      uri = URI("https://#{@host}:#{@port}/api/3/sites/#{site_id}/site_credentials")

      begin
        delete(uri)
        @logger.info("Site #{site_id} credentials removed successfully")
      rescue Exception => e
        @logger.error("Failed to remove Site #{site_id} credentials; #{e}")
      end
    end

    def update_site_credentials(site_id, credentials=[])
      uri = URI("https://#{@host}:#{@port}/api/3/sites/#{site_id}/site_credentials")

      begin
        put(uri, credentials)
        @logger.info("Site #{site_id} updated successfully with #{credentials.length} credentials")
      rescue Exception => e
        @logger.error("Failed to update Site #{site_id} credentials; #{e}")
      end
    end

    def create_site_credential(site_id, credential)
      uri = URI("https://#{@host}:#{@port}/api/3/sites/#{site_id}/site_credentials")

      begin
        post(uri, credential)
        @logger.info("Site #{site_id} credential successfully created")
      rescue Exception => e
        @logger.error("Failed to create Site #{site_id} credential #{credential[:name]}; #{e}")
      end
    end
  end

  class ThycoticOperations
    require 'savon'
    attr_accessor :client
    def initialize(url = nil, comment = '', show_deleted = false, show_restricted = true)
      # log: true, log_level: :info
      @client = Savon.client(wsdl: url, ssl_verify_mode: :none)

      # Comment used when retrieving passwords
      @comment = comment
      @show_deleted = show_deleted
      @show_restricted = show_restricted
    end

    def operations
      puts @client.operations
    end

    def check_type(type)
      case type
        when /.*unix.*/i then 'ssh'
        when /.*windows.*/i then 'cifs'
        when /.*ftp.*/i then 'ftp'
        when /.*400.*/i then 'as400'
        when /.*lotus.*/i then 'notes'
        when /.*Microsoft.*SQL.*Server.*/i then 'tds'
        when /.*Sybase.*SQL.*Server.*/i then 'sybase'
        when /.*mysql.*/i then 'mysql'
        when /.*DB2.*/i then 'db2'
        when /.*postgresql.*/i then 'postgresql'
        when /.*pop.*/i then 'pop'
        when /.*Simple.*Network.*Management.*/i then 'snmp'
        when /.*telnet.*/i then 'telnet'
      end
    end

    def authenticate(username, password)
      operation = :authenticate
      message = { username: username, password: password }
      auth_result = get_secret_result(operation, message)
      check_for_errors(auth_result)
      @token = auth_result[:token]
    end

    def parse_field(secret_response_result, field_name)
      items = secret_response_result[:secret][:items][:secret_item]
      item = items.find  { |i| i[:field_display_name].casecmp(field_name) == 0 }
      item[:value]
    end

    def get_secret_result(operation, message)
      secret = @client.call(operation, message: message)
      resp = secret.hash[:envelope][:body]["#{operation}_response".to_sym]
      resp["#{operation}_result".to_sym]
    end

    def get_secret_summaries(token, ip)
      operation = :search_secrets_by_field_value
      message = {
        token: token,
        fieldName: 'machine',
        searchTerm: ip,
        showDeleted: @show_deleted,
        showRestricted: @show_restricted
      }
      secret_result = get_secret_result(operation, message)

      secrets = []
      unless secret_result[:secret_summaries].nil?
        summaries = secret_result[:secret_summaries][:secret_summary]

        # Ensure summaries is iterable
        summaries = [summaries] if summaries.is_a?(Hash)

        summaries.each do |secret|
          secret_info = {
            secret_id: secret[:secret_id],
            secret_type: secret[:secret_type_name],
            secret_name: secret[:secret_name]
          }
          secrets << secret_info
        end
      end

      secrets
    end

    def get_secret(token, secret_id)
      operation = :get_secret
      message = { token: token,
                  secretId: secret_id,
                  loadSettingsAndPermissions: false,
                  "codeResponses" => {"CodeResponse" =>
                                          [{
                                               "ErrorCode" => "COMMENT",
                                               "Comment" => @comment
                                           }]
                  }}
      secret_result = get_secret_result(operation, message)

      check_for_errors(secret_result)
      username = parse_field(secret_result, 'Username')
      password = parse_field(secret_result, 'Password')

      { username: username, password: password }
    end

    def check_for_errors(result)
      errors = result[:errors]
      unless errors.blank?
        puts errors
        #TODO: Logging
        raise Exception.new(errors)
      end
    end
  end
end
